@@ -20,14 +20,20 @@
 #include "item.h"
 #include "sql_parse.h" // For check_stack_overrun
 
-#ifndef DBUG_OFF
-static int dbug_json_check_min_stack_requirement()
-{
-  my_error(ER_STACK_OVERRUN_NEED_MORE, MYF(ME_FATAL),
-           my_thread_stack_size, my_thread_stack_size, STACK_MIN_SIZE);
-  return 1;
-}
-#endif
+/*
+  Allocating memory and *also* using it (reading and
+  writing from it) because some build instructions cause
+  compiler to optimize out stack_used_up. Since alloca()
+  here depends on stack_used_up, it doesnt get executed
+  correctly and causes json_debug_nonembedded to fail
+  ( --error ER_STACK_OVERRUN_NEED_MORE does not occur).
+*/
+#define ALLOCATE_MEM_ON_STACK(A) do \
+                              { \
+                                uchar *array= (uchar*)alloca(A); \
+                                bzero(array, A); \
+                                my_checksum(0, array, A); \
+                              } while(0)
 
 /*
   Compare ASCII string against the string with the specified
@@ -145,8 +151,11 @@ int json_path_parts_compare(
   int res, res2;
 
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  return dbug_json_check_min_stack_requirement(););
-
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
     return 1;
 
@@ -251,10 +260,6 @@ static int json_nice(json_engine_t *je, String *nice_js,
   static const char *comma= ", ", *colon= "\": ";
   uint comma_len, colon_len;
   int first_value= 1;
-  int value_size = 0;
-  int curr_state= -1;
-  int64_t value_len= 0;
-  String curr_str{};
 
   nice_js->length(0);
   nice_js->set_charset(je->s.cs);
@@ -281,7 +286,6 @@ static int json_nice(json_engine_t *je, String *nice_js,
 
   do
   {
-    curr_state= je->state;
     switch (je->state)
     {
     case JST_KEY:
@@ -329,23 +333,17 @@ handle_value:
         if (append_simple(nice_js, je->value_begin,
                           je->value_end - je->value_begin))
           goto error;
-        
-        curr_str.copy((const char *)je->value_begin,
-                      je->value_end - je->value_begin, je->s.cs);
-        value_len= je->value_end - je->value_begin;
+
         first_value= 0;
-        if (value_size != -1)
-          value_size++;
       }
       else
       {
         if (mode == Item_func_json_format::DETAILED &&
-            depth > 0 && !(curr_state != JST_KEY) &&
+            depth > 0 &&
             append_tab(nice_js, depth, tab_size))
           goto error;
         nice_js->append((je->value_type == JSON_VALUE_OBJECT) ? "{" : "[", 1);
         first_value= 1;
-        value_size= (je->value_type == JSON_VALUE_OBJECT) ? -1: 0;
         depth++;
       }
 
@@ -354,27 +352,11 @@ handle_value:
     case JST_OBJ_END:
     case JST_ARRAY_END:
       depth--;
-      if (mode == Item_func_json_format::DETAILED && (value_size > 1 || value_size == -1) &&
+      if (mode == Item_func_json_format::DETAILED &&
           append_tab(nice_js, depth, tab_size))
         goto error;
-        
-      if (mode == Item_func_json_format::DETAILED && 
-          value_size == 1 && je->state != JST_OBJ_END)
-      {
-        for (auto i = 0; i < value_len; i++)
-        {
-          nice_js->chop();
-        }
-        for (auto i = 0; i < (depth + 1) * tab_size + 1; i++)
-        {
-          nice_js->chop();
-        }
-        nice_js->append(curr_str);
-      }
-      
       nice_js->append((je->state == JST_OBJ_END) ? "}": "]", 1);
       first_value= 0;
-      value_size= -1;
       break;
 
     default:
@@ -536,6 +518,66 @@ longlong Item_func_json_valid::val_int()
 }
 
 
+bool Item_func_json_equals::fix_length_and_dec()
+{
+  if (Item_bool_func::fix_length_and_dec())
+    return TRUE;
+  set_maybe_null();
+  return FALSE;
+}
+
+
+longlong Item_func_json_equals::val_int()
+{
+  longlong result= 0;
+
+  String a_tmp, b_tmp;
+
+  String *a= args[0]->val_json(&a_tmp);
+  String *b= args[1]->val_json(&b_tmp);
+
+  DYNAMIC_STRING a_res;
+  if (init_dynamic_string(&a_res, NULL, 0, 0))
+  {
+    null_value= 1;
+    return 1;
+  }
+
+  DYNAMIC_STRING b_res;
+  if (init_dynamic_string(&b_res, NULL, 0, 0))
+  {
+    dynstr_free(&a_res);
+    null_value= 1;
+    return 1;
+  }
+
+  if ((null_value= args[0]->null_value || args[1]->null_value))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  if (json_normalize(&a_res, a->ptr(), a->length(), a->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  if (json_normalize(&b_res, b->ptr(), b->length(), b->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  result= strcmp(a_res.str, b_res.str) ? 0 : 1;
+
+end:
+  dynstr_free(&b_res);
+  dynstr_free(&a_res);
+  return result;
+}
+
+
 bool Item_func_json_exists::fix_length_and_dec()
 {
   if (Item_bool_func::fix_length_and_dec())
@@ -750,7 +792,7 @@ bool Item_func_json_unquote::fix_length_and_dec()
 {
   collation.set(&my_charset_utf8mb3_general_ci,
                 DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
-  max_length= args[0]->max_char_length() * collation.collation->mbmaxlen;
+  max_length= args[0]->max_length;
   set_maybe_null();
   return FALSE;
 }
@@ -816,7 +858,7 @@ static int alloc_tmp_paths(THD *thd, uint n_paths,
   {
     if (*tmp_paths == 0)
     {
-      MEM_ROOT *root= thd->active_stmt_arena_to_use()->mem_root;
+      MEM_ROOT *root= thd->stmt_arena->mem_root;
 
       *paths= (json_path_with_flags *) alloc_root(root,
           sizeof(json_path_with_flags) * n_paths);
@@ -848,47 +890,21 @@ static void mark_constant_paths(json_path_with_flags *p,
 }
 
 
-Item_json_str_multipath::~Item_json_str_multipath()
+bool Item_json_str_multipath::fix_fields(THD *thd, Item **ref)
 {
-  if (tmp_paths)
-  {
-    for (uint i= n_paths; i>0; i--)
-      tmp_paths[i-1].free();
-  }
+  return alloc_tmp_paths(thd, get_n_paths(), &paths, &tmp_paths) ||
+         Item_str_func::fix_fields(thd, ref);
 }
 
 
-bool Item_json_str_multipath::fix_fields(THD *thd, Item **ref)
+void Item_json_str_multipath::cleanup()
 {
-  if (!tmp_paths)
+  if (tmp_paths)
   {
-    /*
-      Remember the number of paths and allocate required memory on first time
-      the method fix_fields() is invoked. For prepared statements the method
-      fix_fields can be called several times for the same item because its
-      clean up is performed every item a prepared statement finishing its
-      execution. In result, the data member fixed is reset and the method
-      fix_field() is invoked on next time the same prepared statement be
-      executed. On the other side, any memory allocations on behalf of
-      the prepared statement must be performed only once on its first execution.
-      The data member tmp_path is kind a guard to do these activities only once
-      on first time the method fix_field() is called.
-    */
-    n_paths= get_n_paths();
-
-    if (alloc_tmp_paths(thd, n_paths, &paths, &tmp_paths))
-      return true;
+    for (uint i= get_n_paths(); i>0; i--)
+      tmp_paths[i-1].free();
   }
-
-#ifdef PROTECT_STATEMENT_MEMROOT
-  /*
-   Check that the number of paths remembered on first run of a statement
-   never changed later.
-  */
-  DBUG_ASSERT(n_paths == get_n_paths());
-#endif
-
-  return Item_str_func::fix_fields(thd, ref);
+  Item_str_func::cleanup();
 }
 
 
@@ -1021,7 +1037,7 @@ String *Item_func_json_extract::read_json(String *str,
     if (!possible_multiple_values)
     {
       /* Loop to the end of the JSON just to make sure it's valid. */
-      while (json_scan_next(&je) == 0) {}
+      while (json_get_path_next(&je, &p) == 0) {}
       break;
     }
   }
@@ -1151,7 +1167,7 @@ my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
       case JSON_VALUE_OBJECT:
       case JSON_VALUE_ARRAY:
       case JSON_VALUE_FALSE:
-      case JSON_VALUE_UNINITALIZED:
+      case JSON_VALUE_UNINITIALIZED:
       case JSON_VALUE_NULL:
         int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
         return to;
@@ -1197,7 +1213,11 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
   json_engine_t loc_js;
   bool set_js;
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  return dbug_json_check_min_stack_requirement(););
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
     return 1;
 
@@ -1411,19 +1431,10 @@ return_null:
 
 bool Item_func_json_contains_path::fix_fields(THD *thd, Item **ref)
 {
-  /*
-    See comments on Item_json_str_multipath::fix_fields regarding
-    the aim of the condition 'if (!tmp_paths)'.
-  */
-  if (!tmp_paths)
-  {
-    if (alloc_tmp_paths(thd, arg_count-2, &paths, &tmp_paths) ||
-        (p_found= (bool *) alloc_root(thd->active_stmt_arena_to_use()->mem_root,
-                                       (arg_count-2)*sizeof(bool))) == NULL)
-      return true;
-  }
-
-  return Item_int_func::fix_fields(thd, ref);
+  return alloc_tmp_paths(thd, arg_count-2, &paths, &tmp_paths) ||
+         (p_found= (bool *) alloc_root(thd->mem_root,
+                                       (arg_count-2)*sizeof(bool))) == NULL ||
+         Item_int_func::fix_fields(thd, ref);
 }
 
 
@@ -1436,7 +1447,8 @@ bool Item_func_json_contains_path::fix_length_and_dec()
   return Item_bool_func::fix_length_and_dec();
 }
 
-Item_func_json_contains_path::~Item_func_json_contains_path()
+
+void Item_func_json_contains_path::cleanup()
 {
   if (tmp_paths)
   {
@@ -1444,6 +1456,7 @@ Item_func_json_contains_path::~Item_func_json_contains_path()
       tmp_paths[i-1].free();
     tmp_paths= 0;
   }
+  Item_int_func::cleanup();
 }
 
 
@@ -1648,7 +1661,7 @@ bool is_json_type(const Item *item)
     if (Type_handler_json_common::is_json_type_handler(item->type_handler()))
       return true;
     const Item_func_conv_charset *func;
-    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item->real_item())))
+    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item)))
       return false;
     item= func->arguments()[0];
   }
@@ -1783,23 +1796,7 @@ bool Item_func_json_array::fix_length_and_dec()
     return TRUE;
 
   for (n_arg=0 ; n_arg < arg_count ; n_arg++)
-  {
-    ulonglong arg_length;
-    Item *arg= args[n_arg];
-
-    if (arg->result_type() == STRING_RESULT &&
-        !Type_handler_json_common::is_json_type_handler(arg->type_handler()))
-      arg_length= arg->max_char_length() * 2; /*escaping possible */
-    else if (arg->type_handler()->is_bool_type())
-      arg_length= 5;
-    else
-      arg_length= arg->max_char_length();
-
-    if (arg_length < 4)
-      arg_length= 4; /* can be 'null' */
-
-    char_length+= arg_length + 4;
-  }
+    char_length+= static_cast<ulonglong>(args[n_arg]->max_char_length()) + 4;
 
   fix_char_length_ulonglong(char_length);
   tmp_val.set_charset(collation.collation);
@@ -2187,8 +2184,13 @@ err_return:
 
 static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
 {
+
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  return dbug_json_check_min_stack_requirement(););
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
     return 1;
 
@@ -2527,7 +2529,11 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
                           bool *empty_result)
 {
   DBUG_EXECUTE_IF("json_check_min_stack_requirement",
-                  return dbug_json_check_min_stack_requirement(););
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
   if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
     return 1;
 
@@ -3010,12 +3016,8 @@ bool Item_func_json_insert::fix_length_and_dec()
   for (n_arg= 1; n_arg < arg_count; n_arg+= 2)
   {
     paths[n_arg/2].set_constant_flag(args[n_arg]->const_item());
-    /*
-      In the resulting JSON we can insert the property
-      name from the path, and the value itself.
-    */
-    char_length+= args[n_arg/2]->max_char_length() + 6;
-    char_length+= args[n_arg/2+1]->max_char_length() + 4;
+    char_length+=
+        static_cast<ulonglong>(args[n_arg+1]->max_char_length()) + 4;
   }
 
   fix_char_length_ulonglong(char_length);
@@ -3829,20 +3831,7 @@ bool Item_func_json_format::fix_length_and_dec()
 {
   decimals= 0;
   collation.set(args[0]->collation);
-  switch (fmt)
-  {
-  case COMPACT:
-    max_length= args[0]->max_length;
-    break;
-  case LOOSE:
-    max_length= args[0]->max_length * 2;
-    break;
-  case DETAILED:
-    max_length= MAX_BLOB_WIDTH;
-    break;
-  default:
-    DBUG_ASSERT(0);
-  };
+  max_length= args[0]->max_length;
   set_maybe_null();
   return FALSE;
 }
@@ -3976,13 +3965,6 @@ int Arg_comparator::compare_e_json_str_basic(Item *j, Item *s)
   return MY_TEST(sortcmp(res1, res2, compare_collation()) == 0);
 }
 
-bool Item_func_json_arrayagg::fix_fields(THD *thd, Item **ref)
-{
-  bool res= Item_func_group_concat::fix_fields(thd, ref);
-  m_tmp_json.set_charset(collation.collation);
-  return res;
-}
-
 
 String *Item_func_json_arrayagg::get_str_from_item(Item *i, String *tmp)
 {
@@ -4134,7 +4116,7 @@ bool Item_func_json_objectagg::add()
     result.append(STRING_WITH_LEN(", "));
 
   result.append('"');
-  st_append_escaped(&result,key);
+  result.append(*key);
   result.append(STRING_WITH_LEN("\":"));
 
   buf.length(0);
@@ -4155,3 +4137,48 @@ String* Item_func_json_objectagg::val_str(String* str)
 }
 
 
+String *Item_func_json_normalize::val_str(String *buf)
+{
+  String tmp;
+  String *raw_json= args[0]->val_str(&tmp);
+
+  DYNAMIC_STRING normalized_json;
+  if (init_dynamic_string(&normalized_json, NULL, 0, 0))
+  {
+    null_value= 1;
+    return NULL;
+  }
+
+  null_value= args[0]->null_value;
+  if (null_value)
+    goto end;
+
+  if (json_normalize(&normalized_json,
+                     raw_json->ptr(), raw_json->length(),
+                     raw_json->charset()))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+  buf->length(0);
+  if (buf->append(normalized_json.str, normalized_json.length))
+  {
+    null_value= 1;
+    goto end;
+  }
+
+end:
+  dynstr_free(&normalized_json);
+  return null_value ? NULL : buf;
+}
+
+
+bool Item_func_json_normalize::fix_length_and_dec()
+{
+  collation.set(&my_charset_utf8mb4_bin);
+  /* 0 becomes 0.0E0, thus one character becomes 5 chars */
+  fix_char_length_ulonglong((ulonglong) args[0]->max_char_length() * 5);
+  set_maybe_null();
+  return FALSE;
+}
