@@ -29,14 +29,6 @@ class THD;
 class my_decimal;
 class sp_condition_value;
 
-/* Types of LOG warnings, used by note_verbosity */
-
-#define NOTE_VERBOSITY_NORMAL             (1U << 0)
-/* Show warnings about keys parts that cannot be used */
-#define NOTE_VERBOSITY_UNUSABLE_KEYS      (1U << 1)
-/* Show warnings in explain for key parts that cannot be used */
-#define NOTE_VERBOSITY_EXPLAIN            (1U << 2)
-
 ///////////////////////////////////////////////////////////////////////////
 
 class Sql_state
@@ -253,7 +245,8 @@ class Sql_condition_identity: public Sql_state_errno_level,
                               public Sql_user_condition_identity
 {
 public:
-  Sql_condition_identity() = default;
+  Sql_condition_identity()
+  { }
   Sql_condition_identity(const Sql_state_errno_level &st,
                          const Sql_user_condition_identity &ucid)
    :Sql_state_errno_level(st),
@@ -313,6 +306,9 @@ protected:
   /** SQL CURSOR_NAME condition item. */
   String m_cursor_name;
 
+  /** SQL ROW_NUMBER condition item. */
+  ulong m_row_number;
+
   Sql_condition_items()
    :m_class_origin((const char*) NULL, 0, & my_charset_utf8mb3_bin),
     m_subclass_origin((const char*) NULL, 0, & my_charset_utf8mb3_bin),
@@ -323,7 +319,8 @@ protected:
     m_schema_name((const char*) NULL, 0, & my_charset_utf8mb3_bin),
     m_table_name((const char*) NULL, 0, & my_charset_utf8mb3_bin),
     m_column_name((const char*) NULL, 0, & my_charset_utf8mb3_bin),
-    m_cursor_name((const char*) NULL, 0, & my_charset_utf8mb3_bin)
+    m_cursor_name((const char*) NULL, 0, & my_charset_utf8mb3_bin),
+    m_row_number(0)
   { }
 
   void clear()
@@ -338,6 +335,7 @@ protected:
     m_table_name.length(0);
     m_column_name.length(0);
     m_cursor_name.length(0);
+    m_row_number= 0;
   }
 };
 
@@ -441,20 +439,19 @@ private:
     @param level    - the error level for this condition
     @param msg      - the message text for this condition
   */
-  Sql_condition(MEM_ROOT *mem_root,
-                const Sql_condition_identity &value,
-                const char *msg)
-   :Sql_condition_identity(value),
-    m_mem_root(mem_root)
+  Sql_condition(MEM_ROOT *mem_root, const Sql_condition_identity &value,
+                const char *msg, ulong current_row_for_warning)
+   : Sql_condition_identity(value), m_mem_root(mem_root)
   {
-    DBUG_ASSERT(mem_root != NULL);
     DBUG_ASSERT(value.get_sql_errno() != 0);
     DBUG_ASSERT(msg != NULL);
     set_builtin_message_text(msg);
+    m_row_number= current_row_for_warning;
   }
 
   /** Destructor. */
-  ~Sql_condition() = default;
+  ~Sql_condition()
+  {}
 
   /**
     Copy optional condition items attributes.
@@ -599,16 +596,6 @@ private:
   */
   bool has_sql_condition(const char *message_str, size_t message_length) const;
 
-  /**
-    Checks if Warning_info contains SQL-condition with the given error id
-
-    @param sql_errno SQL-condition error number
-
-    @return true if the Warning_info contains an SQL-condition with the given
-    error id.
-  */
-  bool has_sql_condition(uint sql_errno) const;
-
   /**
     Reset the warning information. Clear all warnings,
     the number of warnings, reset current row counter
@@ -736,7 +723,7 @@ private:
   void inc_current_row_for_warning() { m_current_row_for_warning++; }
 
   /** Reset the current row counter. Start counting from the first row. */
-  void reset_current_row_for_warning() { m_current_row_for_warning= 1; }
+  void reset_current_row_for_warning(int n) { m_current_row_for_warning= n; }
 
   ulong set_current_row_for_warning(ulong row)
   {
@@ -765,9 +752,8 @@ private:
 
     @return a pointer to the added SQL-condition.
   */
-  Sql_condition *push_warning(THD *thd,
-                              const Sql_condition_identity *identity,
-                              const char* msg);
+  Sql_condition *push_warning(THD *thd, const Sql_condition_identity *identity,
+                              const char* msg, ulong current_row_number);
 
   /**
     Add a new SQL-condition to the current list and increment the respective
@@ -885,8 +871,8 @@ public:
 class ErrConv: public ErrBuff
 {
 public:
-  ErrConv() = default;
-  virtual ~ErrConv() = default;
+  ErrConv() {}
+  virtual ~ErrConv() {}
   virtual LEX_CSTRING lex_cstring() const= 0;
   inline const char *ptr() const
   {
@@ -1089,11 +1075,6 @@ public:
     return m_statement_warn_count;
   }
 
-  uint unsafe_statement_warn_count() const
-  {
-    return m_statement_warn_count;
-  }
-
   /**
     Get the current errno, state and id of the user defined condition
     and return them as Sql_condition_identity.
@@ -1156,9 +1137,6 @@ public:
   bool has_sql_condition(const char *message_str, size_t message_length) const
   { return get_warning_info()->has_sql_condition(message_str, message_length); }
 
-  bool has_sql_condition(uint sql_errno) const
-  { return get_warning_info()->has_sql_condition(sql_errno); }
-
   void reset_for_next_command()
   { get_warning_info()->reset_for_next_command(); }
 
@@ -1177,8 +1155,8 @@ public:
   void inc_current_row_for_warning()
   { get_warning_info()->inc_current_row_for_warning(); }
 
-  void reset_current_row_for_warning()
-  { get_warning_info()->reset_current_row_for_warning(); }
+  void reset_current_row_for_warning(int n)
+  { get_warning_info()->reset_current_row_for_warning(n); }
 
   bool is_warning_info_read_only() const
   { return get_warning_info()->is_read_only(); }
@@ -1209,10 +1187,12 @@ public:
                               const char* sqlstate,
                               Sql_condition::enum_warning_level level,
                               const Sql_user_condition_identity &ucid,
-                              const char* msg)
+                              const char* msg,
+                              ulong current_row_number)
   {
     Sql_condition_identity tmp(sql_errno_arg, sqlstate, level, ucid);
-    return get_warning_info()->push_warning(thd, &tmp, msg);
+    return get_warning_info()->push_warning(thd, &tmp, msg,
+                                            current_row_number);
   }
 
   Sql_condition *push_warning(THD *thd,
@@ -1222,7 +1202,7 @@ public:
                               const char* msg)
   {
     return push_warning(thd, sqlerrno, sqlstate, level,
-                        Sql_user_condition_identity(), msg);
+                        Sql_user_condition_identity(), msg, 0);
   }
   void mark_sql_conditions_for_removal()
   { get_warning_info()->mark_sql_conditions_for_removal(); }
