@@ -104,7 +104,7 @@ typedef struct st_federatedx_share {
   /*
     the primary select query to be used in rnd_init
   */
-  char *select_query;
+  LEX_CSTRING select_query;
   /*
     remote host info, parse_url supplies
   */
@@ -258,7 +258,6 @@ public:
   void stmt_autocommit();
 };
 
-
 /*
   Class definition for the storage engine
 */
@@ -292,12 +291,10 @@ private:
                                       FEDERATEDX_IO_RESULT *result);
   bool create_where_from_key(String *to, KEY *key_info,
                              const key_range *start_key,
-                             const key_range *end_key,
-                             bool records_in_range, bool eq_range);
+                             const key_range *end_key, bool eq_range);
   int stash_remote_error();
 
-  federatedx_txn *get_txn(THD *thd, bool no_create= FALSE);
-
+  static federatedx_txn *get_txn(THD *thd, bool no_create= FALSE);
   static int disconnect(handlerton *hton, MYSQL_THD thd);
   static int savepoint_set(handlerton *hton, MYSQL_THD thd, void *sv);
   static int savepoint_rollback(handlerton *hton, MYSQL_THD thd, void *sv);
@@ -319,7 +316,7 @@ private:
   int real_connect(FEDERATEDX_SHARE *my_share, uint create_flag);
 public:
   ha_federatedx(handlerton *hton, TABLE_SHARE *table_arg);
-  ~ha_federatedx() = default;
+  ~ha_federatedx() {}
   /*
     The name of the index type that will be used for display
     don't implement this method unless you really have indexes
